/*͸*/
/*                                                                  */
/*  Program Name :  PYTHON.C                                        */
/*                                                                  */
/*        Author :  Robert L. Pyron                                 */
/*                  Turning Point Software                          */
/*                  bobp@turningpoint.com                           */
/*                                                                  */
/*       Version :  1.0                                             */
/* Creation Date :  980918                                          */
/* Revision Date :  981002                                          */
/*                                                                  */
/*͵*/
/*                                                                  */
/*   Description : Codewright extension to provide ChromaCoding,    */
/*                 Smart Indenting, and Language Template features  */
/*                 to PYTHON source files being edited using        */
/*                 Codewright.                                      */
/*                                                                  */
/*                 This extension is known to work with Codewright  */
/*                 version 5.1c, 32-bit. It was compiled with       */
/*                 Microsoft Visual C++ 5.0, Service Pack 3.        */
/*                 Other configurations have not been tested.       */
/*                                                                  */
/*                 This PYTHON version was created by modifying     */
/*                 the AWK version that originally came with        */
/*                 Codewright.                                      */
/*                                                                  */
/*        CAVEAT : I do not use templates or smart indenting,       */
/*                 so those features are completely untested.       */
/*                 USE AT YOUR OWN RISK (better yet, fix them       */
/*                 and send the fixes to Premia) -- bobp            */
/*                                                                  */
/*͵*/
/*                                                                  */
/*               Copyright 1991,1992 Premia Corp.                   */
/*                     All rights reserved                          */
/*                                                                  */
/* This source code is provided to the purchaser of the product     */
/* for the purpose of allowing the purchaser to customize and/or    */
/* extend the functionality of the product.  The source code        */
/* remains the property of Premia Corp. with all rights reserved.   */
/* Any modifications to the source code are considered derivative   */
/* works and all rights thereto are reserved to Premia Corp. except */
/* that the purchaser may use the modified product in the same      */
/* permitted manner as the unmodified product.  It is expressly     */
/* prohibited to distribute in any manner any part of the original  */
/* source code, whether in source or object form, without the       */
/* express written permission from Premia Corp.                     */
/*                                                                  */
/* Owner:                                                           */
/*     Premia Corporation                                           */
/*     1075 NW Murray Blvd., Suite 268                              */
/*     Portland OR 97229                                            */
/*;*/





#include <windows.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "exports.h"
#include "cwstart.h"
#define __PYTHON_STRID__
#include "python.h"

 

#define	SINGLEQUOTE			'\''
#define	DOUBLEQUOTE			'"'
#define	BACKSLASH			'\\'
#define	NEWLINE				'\n'
#define	isquotechar(ch)	( (ch)==SINGLEQUOTE || (ch)==DOUBLEQUOTE )
#define	isbackslash(ch)	( (ch)==BACKSLASH )
#define	iscommentchar(ch)	( (ch)=='#' )



#define MAX_BUF_LEN    100


static LPVOID _py_hash_names[3] = { 0 };	/* pointer to the hash tables containing:			*/
														/* index 0 -- all 'PYTHON' keyword names.			*/
														/* index 1 and 2 -- user defined keyword names.	*/
static int _py_hash_sizes[3] = { 47, 47, 47 };

static LPVOID _py_hash_templates	= NIL;	/* pointer to the hash table containing */
														/* an abbreviated name and its template */



/* Define a structure to hold an abbreviation and the template to be
 * expanded when the abbreviation is entered into the buffer.
 * This structure will be the element in the hash table _py_hash_templates.
 */
typedef struct _LanguageTemplate {         
    LPSTR abbrev;
    LPSTR template;
} LanguageTemplate;


/* This string contains all keyword names which will be indented an extra level.
 */
LPSTR python_INDENT_STR = "\377class\377def\377if\377else\377elif\377while\377for\377try\377except\377finally\377";

/*
 * _py_templates[] contains the keyword abbreviation and the template for the
 * abbreviation.  To add a new name, insert it before the NIL entry or
 * call _py_assign_template().
 *
 * The special characters within the template are:
 *
 *     \n - new line
 *     &  - cursor position after template insertion
 *     @  - backspace
 *     \c - insert 'c' literally
 */
LanguageTemplate _py_templates[] = {

	// I don't use templates. 

	{ NIL, NIL },

};

/* Define the list of 'python' keyword names to appear in
 * the keyword color if language coloring is enabled. 
 */
LPSTR _py_names[] = {

	// reserved keywords
	"and",		"assert",		"break",			"class",
	"continue",	"def",			"del",			"elif",
	"else",		"except",		"exec",			"finally",
	"for",		"from",			"global",		"if",
	"import",	"in",				"is",				"lambda",
	"not",		"or",				"pass",			"print",
	"raise",		"return",		"try",			"while",

	// a few other special identifiers
	"None",		"Ellipsis",
	"self",

    NIL,
};




void DLL	_init();
void DLL	_py_init( void );
BOOL DLL	_py_expand( void );
long DLL	_py_language_colors_update( long firstline, long lastline );
void DLL	_py_language_colors( long numlines );
void DLL	_py_add_keyword( LPSTR keyword, long table );
void DLL	__py_add_keyword( LPSTR keyword, long table, LPVOID *hash_tables, LPINT hash_sizes );
void DLL	_py_assign_template( LPSTR abbrev, LPSTR template );
BOOL DLL	_py_indent( void );
void DLL _py_routines( void );



static BOOL _py_in_comment_or_quote( long offset );
static BOOL	_py_parselongquote( long* starttoken, long* endtoken );
static BOOL	_py_parseshortquote( long* starttoken, long* endtoken );
static BOOL	_py_parsecomment( long* starttoken, long* endtoken );
static BOOL	_py_parsenumber( long* starttoken, long* endtoken );
static BOOL	_py_parsealpha( char* buf, int maxbuf, long* starttoken, long* endtoken );
static BOOL	_py_isfunctioncall( long offset );



LIBMAIN    /* defines the LibMain() */

/*----------------------------------------------------------------------
 ** _init()
 *
 *  Standard initialization function for Codewright extensions.
 *
 *----------------------------------------------------------------------
 */
void DLL 
_init()
{
    _py_init();
}

/*----------------------------------------------------------------------
 ** _py_init()
 *
 *  Initialize this module by exported any required functions and
 *  the hash tables needed to support indenting, template expandion
 *  and keyword coloring.
 *
 *----------------------------------------------------------------------
 */
void DLL
_py_init( void )
{
	static BOOL firstTime = TRUE;
	int    i;

	if (firstTime)
	{
		LibExport( "BOOL _py_expand");
		LibExport( "_py_assign_template LPSTR LPSTR");
		LibExport( "_py_add_keyword LPSTR long");
		LibExport( "BOOL _py_indent");
		LibExport( "_py_language_colors_update long long");
		LibExport( "_py_language_colors long");
		LibExport( "_py_routines" );


		firstTime = FALSE;

		/* 
		 * Create a hash table to hold all of the 'python' keyword names.
		 */
		{
			LPSTR str;
			for ( i = 0; (str = _py_names[ i ]) != NIL; i++ )
				__py_add_keyword( str, 0, _py_hash_names, _py_hash_sizes );
		}

		/* 
		 * Create a hash table to hold all of the 'python' template names.
		 */
		_py_hash_templates = HashCreateTable( 301 /* a prime # */, sizeof( LPSTR ), HASH_IGN_CASE );
		i = 0;
		while ( _py_templates[ i ].abbrev != NIL )
		{
			_py_assign_template( _py_templates[i].abbrev, _py_templates[i].template );
			i++;
		}
	}
}




/*----------------------------------------------------------------------
 ** _py_expand()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'python' language template expansion.  This function
 *    is normally bound to the <SPACE> key.  When invoked, it reads the 
 *    previous word.  If the word is in the list of defined templates,
 *    the template is extracted and inserted.  Some characters in the 
 *    templates have special meanings. They are:
 *
 *       \n - new line
 *       &  - cursor position after template insertion
 *       @  - backspace
 *       \  - take next character literally
 *
 *----------------------------------------------------------------------
 */
BOOL DLL
_py_expand( void )
{
	LPVOID htEntry;
	LPSTR  _py_string = NULL;
	LPSTR  str;
	BOOL   tabUsage;

	MarkSavePos();

	tabUsage = BufQTabUsage();
	BufSetTabUsage( FALSE );

	/* Check to see if we are currently positioned inside a 'python' comment
	 */
	if ( _py_in_comment_or_quote(-1L) )
		goto ___py_expand_incomplete;

	str = BufReadStr( 65535 );     /* read entire line */

	_py_string = StrTrim( str, " \t" );
	StrFree( str );

	if (!_py_string)
		goto ___py_expand_incomplete;

	/* Search for a keyword match in the hash table.
	 */
	htEntry = HashFindEntry( _py_hash_templates, _py_string, strlen( _py_string ) );
	StrFree( _py_string );

	if (!htEntry)
		goto ___py_expand_incomplete;

	/* A template exists for the current keyword, delete the keyword
	 * and insert the template.
	 */
	MarkDropPos( 1 );
	str = *(LPSTR*)htEntry;
	BufDelToEOL();
	_ext_expand_template( str );


	BufSetTabUsage( tabUsage );
	return TRUE;

___py_expand_incomplete:
	MarkRestorePos();
	BufSetTabUsage( tabUsage );

	return FALSE;
}

/*----------------------------------------------------------------------
 ** _py_add_keyword
 *
 *
 *----------------------------------------------------------------------
 */
void DLL
_py_add_keyword( LPSTR keyword, long table )
{
	__py_add_keyword( keyword, table, _py_hash_names, _py_hash_sizes );
}

void DLL
__py_add_keyword( LPSTR keyword, long table, LPVOID *hash_tables, LPINT hash_sizes )
{
	LPSTR XFAR*	lpf	= NULL;
	LPVOID  		ht		= NULL;

	if (keyword && *keyword && table >= 0 && table < 3)
	{
		if ((ht = hash_tables[table]) == NIL)
		{
			hash_tables[table] = ht =
					HashCreateTable( hash_sizes[table], sizeof(LPSTR), 0 );
		}
		if (ht)
			lpf = (LPSTR XFAR *)HashGetEntry( ht, keyword, strlen(keyword) );
		if (lpf)
			*lpf = StrNew( keyword );
	}
}





/*----------------------------------------------------------------------
 ** _py_assign_template()
 *
 *  PARAMETERS:
 *    abbrev   - the keyword abreviation used to locate a template
 *    template - the encoded template string associated with 'abbrev'
 *
 *  DESCRIPTION:
 *    Provide a means of adding or modifying 'python' template strings
 *    without changing the code.
 *
 *----------------------------------------------------------------------
 */
void DLL
_py_assign_template( LPSTR abbrev, LPSTR template )
{
    _ext_assign_template( _py_hash_templates, abbrev, template );
}


/*----------------------------------------------------------------------
 ** _py_indent()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'python' language smart indenting.  This function
 *    will behave differently based on the following conditions:
 *
 *    1. Within a comment - normal indentation
 *
 *    2. Not positioned at end of line - normal indentation
 *
 *    3. Current line prefixed by a name in python_INDENT_STR - perform smart
 *       indentation.  
 *
 *----------------------------------------------------------------------
 */
BOOL DLL
_py_indent( void )
{
	long   curline;
	long   col;
	LPSTR  _py_string;
	LPSTR  str;
	long   matchpos;

	MarkSavePos();

	curline = BufQCurrentLine();

	/* Check to see if we are currently positioned inside a 'python' comment
	*/
	if ( _py_in_comment_or_quote(-1L) )
		goto ___py_indent_incomplete;

	/* Not in a comment so indent according to the current 'python' construct.
	*/
	if (!SrchFind( "^[ \t]*[a-zA-Z_]+\\c([ \t]|$)", SEARCH_REGEX , NIL )
		|| (BufQCurrentLine() != curline) )
		goto ___py_indent_incomplete;

	/* Read the keyword at the beginning of the current line.
	*/
	col = BufQCurrentCol();
	SrchFind( "(^|[^a-zA-Z_])\\c[a-zA-Z_]", SEARCH_REGEX , NIL );
	str = BufReadStr( col - BufQCurrentCol() );
	_py_string = StrTrim( str, " \t" );
	StrFree( str );

	/* Set prefix and suffix to \377 to perform a quick lookup of the name
	*/
	str = StrNew( "\377" );
	_py_string = StrAppend( _py_string, str );
	str = StrAppend( str, _py_string );
	StrFree( _py_string );

	matchpos = StrMatch( str, python_INDENT_STR, SEARCH_IGCASE | SEARCH_FORWARD, NIL );
	StrFree( str );

	if (!matchpos)
	{
		/* Not a 'python' keyword, do normal indentation. */
		goto ___py_indent_incomplete;
	}

	MarkRestorePos();

	_ext_insert_indented_EOL();
	BufInsertChar( '\t' );
	return TRUE;

___py_indent_incomplete:
	MarkRestorePos();

	return FALSE;
}



/*----------------------------------------------------------------------
 ** _py_language_colors_update()
 *
 *  This function does an intermediate update of the colors within
 *  a range of line numbers (firstline and lastline).  It must also
 *  determine if the first and last line specified are within a
 *  comment and act appropriately.
 *
 *----------------------------------------------------------------------
 */
long DLL
_py_language_colors_update( long firstline, long lastline )
{
	long startoff;
	long count = 0;

	if (firstline && lastline)
	{
		/* Begin coloring at the previous comment if
		 * currently located in the middle of one.
		 */
		_PosInit( 0 );
		if (firstline == 1)
		   _PosPrevLine( 0L );
		else
		   _PosNextLine( firstline - 1);
		startoff = _PosQOffset();

		/* Clear all attributes for the line range and
		 * then reset them within _py_language_colors.
		 */
		AttrSetColor( firstline, lastline, 1L, 0x7FFFFFFFL, 0 );
		_PosInit( startoff );
		count = lastline - firstline + 1;
	  _py_language_colors( count );
	}

	return count;
    
}


/*----------------------------------------------------------------------
 ** _py_language_colors
 *
 *
 *----------------------------------------------------------------------
 */
void DLL
_py_language_colors( long numlines )
{
	WORD  ch;
	long	oldoffset;
	long	curoffset;
	long	startoffset;
	long	endoffset;
	long	startline;
	long	endline;
	long	kMaxLine = 0x7fffffffL;

	char	buf[ MAX_BUF_LEN+1 ];
	int   i = 0;
	int   comment_color		= ColorComments( -1 );
	int   keyword_color		= ColorKeywords( -1 );
	long  maxlines 			= ExtCommentSearchLimit( -1 );
	int   quote_color			= ColorAlternate1( -1 );
	int   number_color		= ColorAlternate2( -1 );
	int   brace_color			= ColorAlternate3( -1 );
	int   preproc_color		= ColorAlternate4( -1 );
	int   longquote_color	= comment_color;
	

	// Save the current offset.
	oldoffset = _PosQOffset();

	// Start at the top of the buffer, or at current position.
	switch (numlines)
	{
		case -1L:
			_PosInit( 0 );  
			numlines = kMaxLine;
			break;
		case 0:
			numlines = 1;
			break;
	}

	// Move to start of first line; fetch offset and line number.
	_PosPrevLine(0);
	startoffset = _PosQOffset();
	startline = _PosQLine();

	// Move to end of last line; fetch offset and line number.
	_PosNextLine(numlines);
	endoffset = _PosQOffset();
	endline = _PosQLine();

	// Start at beginning of buffer.
	_PosInit(0);
	curoffset = _PosQOffset();
	ch = _PosCurrentChar();
	while ( ch != EOF_CHAR  &&  curoffset < endoffset )
	{
		// Note that "endtoken", when valid, will be set to offset
		// of last character of the token.
		long	starttoken, endtoken;
		long	tokencount;

		// We don't start coloring until we reach startoffset.
		// The only token types that can cross line boundaries
		// are long quotes and short quotes, so we can ignore
		// everything up to the first quote or comment character.
		while ( curoffset < startoffset )
		{
			if ( isquotechar(ch) || iscommentchar(ch) )
				break;
			ch = _PosNextChar();
			curoffset = _PosQOffset();
		}

		// Deal with long quotes ("""...""" or '''...''')
		if ( _py_parselongquote(&starttoken, &endtoken) )
		{
			tokencount = endtoken - starttoken + 1;
			if ( endtoken < startoffset )		// don't start coloring until we reach startoffset
				;
			else if ( longquote_color  &&  quote_color  &&  tokencount > 6 )
			{
				long	startpos, endpos;

				// opening quotation marks
				startpos = max(starttoken,startoffset);
				endpos = min(endtoken+1,endoffset);			// one beyond end of token
				if ( endpos > startpos )
				{
					_PosInit( startpos );
					_PosSetColor( quote_color, endpos-startpos );
				}

				// quoted text
				startpos = max(starttoken+3,startoffset);
				endpos = min(endtoken+1,endoffset);			// one beyond end of token
				if ( endpos > startpos )
				{
					_PosInit( startpos );
					_PosSetColor( longquote_color, endpos-startpos );
				}

				// closing quotation marks
				startpos = max(endtoken-2,startoffset);
				endpos = min(endtoken+1,endoffset);			// one beyond end of token
				if ( endpos > startpos )
				{
					startpos = max(startpos,startoffset);
					endpos = min(endpos,endoffset+1);
					_PosInit( startpos );
					_PosSetColor( quote_color, endpos-startpos );
				}
			}
			else if ( quote_color )
			{
				long startpos = max(starttoken,startoffset);
				long endpos = min(endtoken+1,endoffset);	// one beyond end of token
				if ( endpos > startpos )
				{
					_PosInit( startpos );
					_PosSetColor( quote_color, endpos-startpos );
				}
			}
			_PosInit( endtoken );
		}

		// Deal with short quotes ("..." or '...')
		else if ( _py_parseshortquote(&starttoken, &endtoken) )
		{
			tokencount = endtoken - starttoken + 1;
			if ( quote_color  &&  endtoken > startoffset )
			{
				long startpos = max(starttoken,startoffset);
				long endpos = min(endtoken+1,endoffset);	// one beyond end of token
				if ( endpos > startpos )
				{
					_PosInit( startpos );
					_PosSetColor( quote_color, endpos-startpos );
				}
			}
			_PosInit( endtoken );
		}

		// Deal with comments (# to end of line).
		else if ( _py_parsecomment(&starttoken, &endtoken) )
		{
			tokencount = endtoken - starttoken + 1;
			if ( comment_color  &&  endtoken > startoffset )
			{
				long startpos = max(starttoken,startoffset);
				long endpos = min(endtoken+1,endoffset);	// one beyond end of token
				if ( endpos > startpos )
				{
					_PosInit( startpos );
					_PosSetColor( comment_color, endpos-startpos );
				}
			}
			_PosInit( endtoken );
		}

		// We don't start coloring until we reach startoffset.
		else if ( curoffset < startoffset )
		{
			ch = _PosNextChar();
			curoffset = _PosQOffset();
			continue;
		}

		// Deal with numbers.
		else if ( _py_parsenumber(&starttoken, &endtoken) )
		{
			tokencount = endtoken - starttoken + 1;
			if ( number_color )
			{
				_PosInit( starttoken );
				_PosSetColor( number_color, tokencount );
			}
			_PosInit( endtoken );
		}


		// Deal with keywords and function calls.
		else if ( _py_parsealpha( buf, sizeof(buf)-1, &starttoken, &endtoken) )
		{
			int color = 0;
			tokencount = endtoken - starttoken + 1;

			// Note that C language chromacoding gives precedence to
			// table 1 over table 2 and table 0 (see CWSTART\C.C).

			if (_py_hash_names[2] && HashFindEntry( _py_hash_names[2], buf, tokencount ))
				color = ColorAlternate6( -1 );
			else if (_py_hash_names[1] && HashFindEntry( _py_hash_names[1], buf, tokencount ))
				color = ColorAlternate5( -1 );
			else if (HashFindEntry( _py_hash_names[0], buf, tokencount ))
				color = ColorKeywords( -1 );
			else if ( _py_isfunctioncall(endtoken+1) )
				color = ColorAlternate7( -1 );

			if ( color )
			{
				_PosInit( starttoken );
				_PosSetColor( color, tokencount );
			}

			_PosInit( endtoken );
		}

		// Check for bad characters.
		else if ( !isascii(ch) || strchr("@$?",ch) )
		{
			int color = ColorError( -1 );
			if ( color )
				_PosSetColor( color, 1 );
		}

		// Check for operators.
		else if ( strchr("+-*/%&|^~!=<>.`:",ch) )
		{
			int color = ColorAlternate8( -1 );
			if ( color )
				_PosSetColor( color, 1 );
		}

		// Check for brackets; use BRACE color.
#if	0
		else if ( strchr("()[]{}",ch) )
#else
		else if ( strchr("[]{}",ch) )
#endif
		{
			int color = ColorAlternate3( -1 );
			if ( color )
				_PosSetColor( color, 1 );
		}

		ch = _PosNextChar();
		curoffset = _PosQOffset();
	}

	// restore current offset
	_PosInit(oldoffset);
}




/*--  Private functions --*/


/*----------------------------------------------------------------------
 ** _py_parselongquote
 *
 *		Determine whether a long quote begins at the current offset.
 *
 *	returns:	
 *
 *		FALSE if not a long quote
 *				_Pos offset is unchanged
 *
 *		TRUE if a long quote
 *				*starttoken is offset of first character
 *				*endtoken is offset of last character
 *				_Pos offset is offset of last character of long quote
 *
 *----------------------------------------------------------------------
 */
static BOOL	
_py_parselongquote( long* starttoken, long* endtoken )
{
	long	oldoffset	= _PosQOffset();
	WORD	ch0, ch1, ch2;
	WORD	quotecharacter;

	*starttoken = *endtoken = _PosQOffset();

	// Fetch three characters, starting at current offset.
	ch0 = _PosCurrentChar();
	ch1 = _PosNextChar();
	ch2 = _PosNextChar();

	if ( !isquotechar(ch0)  ||  ch1!=ch0  ||  ch2!=ch0 )
	{
		// restore current offset
		_PosInit(oldoffset);
		return FALSE;
	}

	// Save quote character.
	quotecharacter = ch0;

	ch0 = _PosNextChar();
	ch1 = _PosNextChar();
	ch2 = _PosNextChar();

	while (TRUE)
	{
		// end of file
		if ( ch0==EOF_CHAR  ||  ch1==EOF_CHAR  ||  ch2==EOF_CHAR )
			break;

		// end of quote
		if ( ch0==quotecharacter  &&  ch1==quotecharacter  &&  ch2==quotecharacter )
			break;

		if ( ch0 == BACKSLASH )
		{
			ch1 = ch2;
			ch2 = _PosNextChar();
		}

		ch0 = ch1;
		ch1 = ch2;
		ch2 = _PosNextChar();
	}

	*endtoken = _PosQOffset();
	return TRUE;
}



/*----------------------------------------------------------------------
 ** _py_parseshortquote
 *
 *		Determine whether a short quote begins at the current offset.
 *
 *	returns:	
 *
 *		FALSE if not a short quote
 *				_Pos offset is unchanged
 *
 *		TRUE if a short quote
 *				*starttoken is offset of first character
 *				*endtoken is offset of last character
 *				_Pos offset is offset of last character of short quote
 *
 *----------------------------------------------------------------------
 */
static BOOL	
_py_parseshortquote( long* starttoken, long* endtoken )
{
	long	oldoffset	= _PosQOffset();
	WORD	ch				= _PosCurrentChar();
	WORD	quotecharacter;

	*starttoken = *endtoken = _PosQOffset();

	// Fetch characters at current offset.
	quotecharacter = _PosCurrentChar();
	if ( !isquotechar(quotecharacter) )
		return FALSE;

	// Scan until we find a matching quote character.
	ch = _PosNextChar();
	while ( ch!=EOF_CHAR  &&  ch!=quotecharacter )
	{
		if (ch == BACKSLASH)
			_PosNextChar();		// skip escaped character
		ch = _PosNextChar();
	}

	*endtoken = _PosQOffset();
	return TRUE;
}


/*----------------------------------------------------------------------
 ** _py_parsecomment
 *
 *		Determine whether a comment begins at the current offset.
 *
 *	returns:	
 *
 *		FALSE if not a comment
 *				_Pos offset is unchanged
 *
 *		TRUE if a comment
 *				*starttoken is offset of first character
 *				*endtoken is offset of last character
 *				_Pos offset is offset of last character of comment
 *
 *----------------------------------------------------------------------
 */
static BOOL	
_py_parsecomment( long* starttoken, long* endtoken )
{
	long	oldoffset	= _PosQOffset();
	WORD	ch				= _PosCurrentChar();

	*starttoken = *endtoken = _PosQOffset();

	ch = _PosCurrentChar();
	if (ch != '#')
		return FALSE;

	// Scan until we find end of line.
	while (TRUE)
	{
		ch = _PosNextCharNL();
		if ( ch == EOF_CHAR )
			break;
		else if (ch == EOL1_CHAR)
			break;
		else if (ch == EOL2_CHAR)
			break;
		else if (ch == BACKSLASH)
			ch = _PosNextCharNL();		// skip escaped character
	}
	
	*endtoken = _PosQOffset();
	return TRUE;
}

/*----------------------------------------------------------------------
 ** _py_parsenumber
 *
 *		Determine whether a number begins at the current offset.
 *
 *	returns:	
 *
 *		FALSE if not a number
 *				_Pos offset is unchanged
 *
 *		TRUE if a number
 *				*starttoken is offset of first character
 *				*endtoken is offset of last character
 *				_Pos offset is offset of last character of number
 *
 *----------------------------------------------------------------------
 */
static BOOL	
_py_parsenumber( long* starttoken, long* endtoken )
{
	long	oldoffset	= _PosQOffset();
	WORD	ch;
	WORD	lastch;
	int	j;
	BOOL	hex	= FALSE;

	*starttoken = *endtoken = _PosQOffset();

	// check for digit
	ch = _PosCurrentChar();
	if ( !isascii(ch) || !isdigit(ch) )
		return FALSE;

	// Examine previous character. It should be whitespace or punctuation.
	lastch = _PosPrevChar();
	if (lastch != EOF_CHAR)
		_PosNextChar();
	switch (lastch)
	{
		case ' '     :
		case '\t'    :
		case '\r'    :
		case '\n'    :   
		case EOF_CHAR:
			break;			// no problem
		case '_'     :
			return FALSE;
		default      :
			if ( isascii(lastch) && ispunct(lastch) )
				break;		// no problem
			return FALSE;
	}
			
	j = 0;
	hex = FALSE;
	while (TRUE)
	{
		if ( !isascii(ch) )
			break;
		else if ( isdigit(ch) )
			;
		else if ( ch == '.' )
			;
		else if ( ch == 'x' && j == 1 )
			hex = TRUE;
		else if ( ch == 'X' && j == 1 )
			hex = TRUE;
		else if ( hex && isxdigit(ch) )
			;
		else
			break;

		ch = _PosNextChar();
		++j;
	}

	if (ch == 'l' || ch == 'L')
		ch = _PosNextChar();

	if (ch == 'j' || ch == 'J')
		ch = _PosNextChar();

	if ( ch != EOF_CHAR )	// if not at end of buffer ...
		_PosPrevChar();		// ... set _Pos offset to last good char

	*endtoken = _PosQOffset();
	return TRUE;
}





/*----------------------------------------------------------------------
 ** _py_parsealpha
 *
 *		Determine whether an alphnumeric token begins at the current offset.
 *
 *	returns:	
 *
 *		FALSE if not alphanumeric
 *				_Pos offset is unchanged
 *
 *		TRUE if a number
 *				*starttoken is offset of first character
 *				*endtoken is offset of last character
 *				_Pos offset is offset of last character of token
 *
 *----------------------------------------------------------------------
 */
static BOOL	
_py_parsealpha( char* buf, int maxbuf, long* starttoken, long* endtoken )
{
	long	oldoffset	= _PosQOffset();
	WORD	ch;
	WORD	lastch;
	int	i;

	*starttoken = *endtoken = _PosQOffset();

	// check for alpha or underscore
	ch = _PosCurrentChar();
	if ( !isascii(ch) || !iscsymf(ch) )		// letter or underscore
		return FALSE;

	// Examine previous character. It should be whitespace or punctuation.
	lastch = _PosPrevChar();
	if (lastch != EOF_CHAR)
		_PosNextChar();
	switch (lastch)
	{
		case ' '     :
		case '\t'    :
		case '\r'    :
		case '\n'    :   
		case EOF_CHAR:
			break;			// no problem
		default      :
			if ( !isascii(lastch) || !ispunct(lastch) )
				return FALSE;
			break;		// no problem
	}

	// Stuff characters into buf[].
	i = 0;
	while ( isascii(ch) && iscsym(ch) )		// letter, underscore, or digit
	{
		if ( i+1 < maxbuf )
			buf[i++] = (BYTE)ch;
		ch = _PosNextChar();
	}
	buf[i] = 0;


	if ( ch != EOF_CHAR )	// if not at end of buffer ...
		_PosPrevChar();		// ... set _Pos offset to last good char

	*endtoken = _PosQOffset();
	return TRUE;
}



/*----------------------------------------------------------------------
 ** _py_isfunctioncall
 *
 *		Determine whether the token beginning at offset represents a function call.
 *
 *	returns:	
 *
 *		TRUE if function call
 *		FALSE if not function call
 *
 *		_Pos offset is unchanged for both cases.
 *
 *----------------------------------------------------------------------
 */
static BOOL	
_py_isfunctioncall( long offset )
{
	WORD	ch;

	_PosInit( offset );

	ch = _PosCurrentChar();
	while ( ch == ' '  ||  ch == '\t' )
		ch = _PosNextChar();

	_PosInit( offset );
	return ( ch == '(' );
}



/*----------------------------------------------------------------------
 ** _py_in_comment_or_quote
 *
 *  Determines whether or not the current position in the buffer is
 *  located within a 'python' comment.
 *
 *  Returns TRUE if in a comment or quote.
 *
 *  NOTE: leaves "_Pos..." cursor in unknown state
 *
 *----------------------------------------------------------------------
 */
static BOOL
_py_in_comment_or_quote( long offset )
{
	WORD  ch;
	long	curoffset;
	long	desiredoffset;
	
	_PosInit( offset );
	desiredoffset = _PosQOffset();


	// Start at beginning of buffer.
	_PosInit( 0 );
	curoffset = _PosQOffset();
	ch = _PosCurrentChar();

	while ( ch != EOF_CHAR  &&  curoffset < desiredoffset )
	{
		long	starttoken, endtoken;

		// The only token types that can cross line boundaries
		// are long quotes and short quotes, so we can ignore
		// everything up to the first quote or comment character.
		while ( curoffset < desiredoffset )
		{
			if ( isquotechar(ch) || iscommentchar(ch) )
				break;
			ch = _PosNextChar();
			curoffset = _PosQOffset();
		}

		// Deal with long quotes ("""...""" or '''...''')
		// Deal with short quotes ("..." or '...')
		// Deal with comments (# to end of line).
		if (   _py_parselongquote(&starttoken, &endtoken)
		    || _py_parseshortquote(&starttoken, &endtoken)
		    || _py_parsecomment(&starttoken, &endtoken)
		   )
		{
			if ( endtoken <= desiredoffset )
			{
				// we haven't reached the desired offset yet; try again
				_PosInit( endtoken );
				curoffset = _PosQOffset();
				ch = _PosCurrentChar();
				continue;
			}
			else if ( starttoken >= desiredoffset )
			{
				// quote starts after the desired offset
				return FALSE;
			}
			else
			{
				return TRUE;
			}

		}

		ch = _PosNextChar();
		curoffset = _PosQOffset();
	}

}



/*----------------------------------------------------------------------
 ** _py_routines
 *
 * Search for Python function definitions.
 *
 *----------------------------------------------------------------------
 */

// RE_CLASS -- regular expression to detect a class definition
//
// class                                       token "class"
//      [ \t]+                                  whitespace
//            [a-zA-Z0-9_]+                      alphanumeric or underscore (at least one character)
//                         [ \t]*                 optional whitespace
//                               (                 (begin optional phrase)
//                                \\(               open parenthesis
//                         	       [^)]*           zero or more characters, except closing parenthesis
//                         	            \\)         close parenthesis
//                                           )         (end optional phrase)
//                                            &         zero or one occurence of optional phrase

#define	RE_CLASS	"class[ \t]+[a-zA-Z0-9_]+[ \t]*(\\([^)]*\\))?"


// RE_DEF -- regular expression to detect a function definition
//
// def                                    token "def"
//   [ \t]+                                whitespace
//         [a-zA-Z0-9_]+                    alphanumeric or underscore (at least one character)
//                      [ \t]*               optional whitespace
//                            \\(             open parenthesis
//                               [^)]*         zero or more characters, except closing parenthesis
//                         	        \\)       close parenthesis

#define	RE_DEF	"def[ \t]+[a-zA-Z0-9_]+[ \t]*\\([^)]*\\)"


char	_py_str[]	= "^[ \t]*(" RE_CLASS ")|(" RE_DEF ")";
char	_py_stre[]	= "^[ \t]*(" RE_CLASS ")|(" RE_DEF ")\\c";


void DLL
_py_routines( void )
{
	int   matches = 0;
	long  curline;
	long  lines;


	//	SysWaitStart signals the beginning of a possibly lengthy operation 
	//	by changing the standard arrow-shaped mouse cursor into an hour glass.
	SysWaitStart();

	//	MarkSavePos saves the current buffer position described by the parameters 
	//	in a mark on a stack.  This stack is called the "saved positions" stack.
	//	The MarkRestorePos function may be used to move to a saved position.
	MarkSavePos();

	//	BufSetAscii sets the current buffer to ASCII (normal text) mode.  This is 
	//	the standard mode of operation.  Other modes of operation include Compact 
	//	mode and Hex-ASCII/Hex-Binary modes.
	BufSetAscii();

	//	MovTopBuf repositions the cursor at the first line of the buffer.  
	//	Codewright will attempt to maintain the column position.  A call to 
	//	MovHome may therefore be required to bring the cursor to the beginning 
	//	of the buffer.
	MovTopBuf();

	//	MovHome moves the cursor to the beginning of the current line.
	MovHome();

	// Find start of class or function definition.
	//	BOOL SrchFind(LPSTR sPattern, DWORD sFlag, LPLong matchLen)
	while (SrchFind( _py_str, (SEARCH_REGEX | SEARCH_FORWARD) & ~SEARCH_MAXIMAL, 0 ))
	{
		//	BufQCurrentLine reports the line number associated with 
		// the current buffer position.
		curline = BufQCurrentLine();

		// Find end of class or function definition.
		//	BOOL SrchFind(LPSTR sPattern, DWORD sFlag, LPLong matchLen)
		SrchFind( _py_stre, (SEARCH_REGEX | SEARCH_FORWARD) & ~SEARCH_MAXIMAL, 0 );

		if (!matches)
		{
			//	void	AttrSetVisible(DWORD first, DWORD last, int visAttr)
			// assigns a visibility attribute to a range of lines in the current buffer.
			// In this case, we know there is a match, so we set all lines
			// invisible. Matched lines will be turned back on as they are encountered.
			AttrSetVisible( 0, 0, 0 );
		}
		matches++;

		//	BufQCurrentLine reports the line number associated with the current 
		//	buffer position.
		lines = BufQCurrentLine() - curline;
		lines = max(lines, 0L);

		//	void	AttrSetVisible(DWORD first, DWORD last, int visAttr)
		// Turn visibility ON for this class or function definition.
		AttrSetVisible( curline, curline + lines, 1 );

		//	MovHome moves the cursor to the beginning of the current line.
		MovHome();

		//	MovDown moves the cursor position in line units, usually toward the 
		//	end of the current buffer.
		if (!MovDown( 1 ))
			break; 
	}

	if (matches)
	{
		//	BufSetCompact sets the current buffer to compact or elided mode.  
		//	In this mode, only lines selected with the AttrSelVisible or 
		//	AttrSetVisible function are displayed.  Compact mode may only be 
		//	turned off by selecting another mode:  ASCII, or one of the two 
		//	Hex modes.
		BufSetCompact();
		//	ERRCODE	MarkDropPos(int count)
		//	MarkDropPos disposes of saved cursor positions, stored on the stack.  
		//	The number of positions removed from the stack is indicated by the 
		//	parameter.
		MarkDropPos( 1 );
		//	MovTopBuf repositions the cursor at the first line of the buffer.  
		//	Codewright will attempt to maintain the column position.  A call to 
		//	MovHome may therefore be required to bring the cursor to the beginning 
		//	of the buffer.
		MovTopBuf();
	}
	else
	{
		//	MarkRestorePos pops the most recently saved cursor position off the 
		//	stack and makes that position the current position.  Cursor positions 
		//	are saved onto the stack through a call to MarkSavePos.
		MarkRestorePos();
	}

	//
	//	SysWaitEnd signals the end of a possibly lengthy operation by 
	//	resetting the mouse cursor from an hour glass to the standard arrow.
	//
	SysWaitEnd();

#if	0
	// display message indicating result
	if (matches)
		MsgMessage("");
	else
		MsgStrMessage(ModuleInstance, MSG_TBL, IDS_NO_ROUTINES, MSG_MESSAGE);
#endif

}


